# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# TODO(stephenfin): Switch additionalProperties to False in a future version
index_request_query = {
    'type': 'object',
    'properties': {},
    'required': [],
    'additionalProperties': True,
}

index_response_body = {
    'type': 'object',
    'properties': {
        'shards': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'count': {'type': 'integer', 'minimum': 1},
                    'name': {'type': 'string'},
                },
                'required': ['count', 'name'],
                'additionalProperties': False,
            },
        },
    },
    'required': ['shards'],
    'additionalProperties': False,
}
