#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add automated_clean field

Revision ID: d2b036ae9378
Revises: 664f85c2f622
Create Date: 2018-07-25 15:30:20.860792

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd2b036ae9378'
down_revision = '664f85c2f622'


def upgrade():
    op.add_column('nodes', sa.Column('automated_clean', sa.Boolean(),
                                     nullable=True))
