# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""oslo.i18n integration module.

See https://docs.openstack.org/oslo.i18n/latest/user/

"""

try:
    import oslo_i18n as i18n
except ImportError:
    def _(msg):
        return msg

    def install(domain):
        # NOTE(dtantsur): this is called before logging is initialized, so we
        # cannot really log a warning here.
        pass
else:
    _translators = i18n.TranslatorFactory(domain='ironic')

    # The primary translation function using the well-known name "_"
    _ = _translators.primary

    install = i18n.install
