# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Mapping of boot modes used when requesting the system to boot
using alternative firmware interfaces.

The options presented were based on the Redfish protocol capabilities,
specifically on the BootSourceOverrideMode property.
"""

LEGACY_BIOS = 'bios'
"Boot over legacy PC BIOS firmware interface"

UEFI = 'uefi'
"Boot over Unified Extensible Firmware Interface (UEFI) firmware interface"
